import hashlib
import string
from random import random
from password_generator import PasswordGenerator
from django.http import HttpResponse
from django.core.validators import validate_email
from django.core.exceptions import ValidationError
from main import default_vm_node, default_iso_storage, default_storage, ignore_vms_ids
from main import proxmoxAuth
from main.VmProxMoxDTO import VmProxMoxDTO
from main.db_functions import get_default_settings, is_iso_in_sql_db, insert_iso, get_list_iso_from_db, remove_iso, \
    insert_member, is_user_in_sql_db, get_list_users_from_db, remove_user_db, is_vm_in_sql_db, insert_vm_with_id, \
    get_list_db_vm, remove_vm_db, select_user_db, add_user_to_vm
from main.UserProxMoxDTO import UserProxMoxDTO
from datetime import datetime, timedelta
import re
from main.models import Iso
import time


def create_vm(vmdto: VmProxMoxDTO, vm_id: int, vlantag: int):
    # Get parameters for VM
    node = vmdto.node  # Node name
    vm_name = re.sub('[^a-zA-Z0-9]', '', vmdto.vm_name)  # Name for VM sanitized of prohibited symbols
    vm_cores = vmdto.vm_cores  # Number of CPU cores
    vm_memory = vmdto.vm_memory  # Memory size in MB
    vm_storage = vmdto.vm_storage  # Storage ID where the VM will be created
    vm_disk_size = vmdto.vm_disk_size  # Disk size for the VM
    iso_file = vmdto.vm_iso_file  # Path to the ISO file

    config = {
        'name': vm_name,
        'cores': vm_cores,
        'memory': vm_memory,
        'storage': vm_storage,
        'scsi0': f'{vm_storage}:{vm_disk_size[:-1]}',  # Disk size
        'boot': 'cdn',
        'cdrom': iso_file,
        'net0':  f'virtio,bridge=vmbr0,tag={vlantag}'
    }
    # Send request to create a VM
    out = proxmoxAuth.nodes(node).qemu.create(vmid=vm_id, **config)
    return out


# ('pve', 'local')
def get_list_iso_urls(node, storage) -> list:
    # Request list of all ISO files
    iso_info = proxmoxAuth.nodes(node).storage(storage).content.get(content='iso')

    # Create a dictionary of iso_name - iso_path
    iso_list_output = []
    for record in iso_info:
        iso_list_output.append(record['volid'])

    return iso_list_output


def get_list_templates() -> list:
    template_info = proxmoxAuth.cluster.resources.get(type='vm')
    templates = []
    for record in template_info:
        try:
            if record['template'] == 1:
                templates.append(record)
        except KeyError:
            continue # If vm not fully initialized -> skip it
    return templates


def get_list_all_users() -> list:
    user_info = proxmoxAuth.access.users.get()
    return user_info


def months_to_days(num_months):
    current_date = datetime.now()
    future_date = current_date + timedelta(days=(num_months * 365) / 12)
    days_difference = (future_date - current_date).days
    return days_difference


def sanitize_vm_data_input(vminp: VmProxMoxDTO) -> VmProxMoxDTO:
    # Get a list of limits from mysql DB
    limits = get_default_settings()
    # Validate input against limits and change data accordingly
    if limits['storage_gbs'] < int(vminp.vm_disk_size[:-1]):  # Remove "G" at the end of disk size
        vminp.vm_disk_size = limits['storage_gbs']
    if limits['memory_mbs'] < vminp.vm_memory:
        vminp.vm_memory = limits['memory_mbs']
    if limits['cores'] < vminp.vm_cores:
        vminp.vm_cores = limits['cores']
    if months_to_days(int(limits['lease_time_months'])) < vminp.vm_expire_days:
        vminp.vm_expire_days = months_to_days(limits['lease_time_months'])
    return vminp


def sync_iso_from_ve_to_db():
    proxmox_isos = get_list_iso_urls(default_vm_node, default_iso_storage)
    for volid in proxmox_isos:
        if not is_iso_in_sql_db(volid):
            # If iso is not in sql then save its data
            insert_iso(volid, volid)
    # If sql has ISO that no longer exists in VE, remove it
    sql_iso = get_list_iso_from_db()
    for record in sql_iso:
        if record.volid not in proxmox_isos:
            remove_iso(record.id)


# Returns volid of the uploaded file
def upload_iso(iso_file, upload_path) -> str:
    # Upload to Proxmox VE
    proxmoxAuth.nodes(default_vm_node).storage(default_iso_storage).upload.create(filename=open(upload_path, 'rb'),
                                                                                  content='iso')
    # Get last uploaded ISO's info for volid
    iso_files = proxmoxAuth.nodes('pve').storage('local').content.get()
    last_uploaded_iso = max(iso_files, key=lambda x: x['ctime'])
    return last_uploaded_iso['volid']


def remove_iso_from_ve(iso_id: str) -> HttpResponse:
    try:
        iso_model = Iso.objects.get(id=iso_id)
    except Iso.DoesNotExist:
        return HttpResponse("Object not found.")

    try:
        # If the object in db is found -> remove
        proxmoxAuth.nodes(default_vm_node).storage(default_iso_storage).content.delete(iso_model.volid)
        return HttpResponse("File removed!")
    except Exception as e:
        return HttpResponse({str(e)})


def create_user(user: UserProxMoxDTO, useridinp: int):
    config = {
        'password': user.password,
        'firstname': user.firstname,
        'lastname': user.lastname,
        'email': user.email,
        'enable': user.enable,
    }
    user_name_realm = str(user.username) + '@' + "pve"
    out = proxmoxAuth.access.users.post(userid=user_name_realm, **config)


def sha256_hash_string(input: str) -> str:
    # Convert the input to bytes
    password_bytes = input.encode('utf-8')
    # Generate the SHA-256 hash
    sha256_hash = hashlib.sha256(password_bytes).hexdigest()
    return sha256_hash


def is_email_nhl(email: str) -> bool:
    try:
        validate_email(email)
        domain = email.split("@")[1]
        if domain == "nhlstenden.com" or domain == "student.nhlstenden.com":
            return True
        else:
            return False
    except ValidationError:
        return False


def change_user_enable(username: str) -> HttpResponse:
    username_formatted = username + '@pve'
    user = proxmoxAuth.access.users(username_formatted).put(enable=1)
    return user


def remove_user_proxmox_ve(username: str) -> HttpResponse:
    userid = username + "@pve"
    response = proxmoxAuth.access.users(userid).delete()
    return response


def sync_users_ve_db():
    proxmox_users = get_list_all_users()
    list_of_usernames = []
    for user in proxmox_users:
        username_prox = str(user['userid']).split('@')[0]
        list_of_usernames.append(username_prox)
        if not is_user_in_sql_db(username_prox):
            # If user is not in sql then save its data
            # Since password cannot be retrieved from proxmox, set password to Username value. User can later change the password
            userdto = UserProxMoxDTO(user['email'], user['firstname'], user['lastname'], username_prox, username_prox)
            insert_member(userdto)
        # If sql has user that no longer exists in VE, remove it
    sql_users = get_list_users_from_db()
    for record in sql_users:
        if record.username not in list_of_usernames:
            remove_user_db(record.username)


def create_vm_from_template(vmid: int, vmname: str, templateid: int, vlantag: int) -> HttpResponse:
    response = proxmoxAuth.nodes(default_vm_node).qemu(templateid).clone.create(newid=vmid, name=vmname, full=1)
    time.sleep(2)
    counter = 0
    # Wait for VM to initialize and get its config
    while True:
        if counter == 100:
            break
        try:
            config = proxmoxAuth.nodes(default_vm_node).qemu(vmid).config.get()
            tempnet = config['net0']
            break
        except:
            time.sleep(1)
            counter = counter + 1
    tempnet = config['net0']
    tempnet = tempnet + f',tag={vlantag}'
    # Update config to include user's vlantag
    config['net0'] = tempnet
    response_upd = proxmoxAuth.nodes(default_vm_node).qemu(vmid).config.post(net0=config['net0'])
    return response_upd


# Check if a logged-in user is a superuser
def superuser_check(user):
    return user.is_superuser


def add_user_to_vm_proxmox(username: str, vmid: int) -> HttpResponse:
    path_temp = '/vms/' + str(vmid)
    user_temp = username + '@pve'
    response = proxmoxAuth.access.acl.put(path=path_temp, roles='PVEVMUser', users=user_temp)
    return response


def get_resource_consumption() -> dict:
    # Get resource consumption information
    response = proxmoxAuth.nodes(default_vm_node).status.get()
    return response


def sync_vms_ve_and_db():
    proxmox_vms = get_list_vms_proxmox()
    default = get_default_settings()
    vm_id_list_prox = []
    for record in proxmox_vms:
        try:
            if record['template'] == 1 or record['vmid'] in ignore_vms_ids:
                continue
        except KeyError:
            continue
        vm_id_list_prox.append(record['vmid'])
        if not is_vm_in_sql_db(record['vmid']):
            # If VM is not in sql then save its data
            vmdto = VmProxMoxDTO(record['name'], default['cores'], default['memory_mbs'], default['storage_gbs'],
                                 months_to_days(default['lease_time_months']), None)
            insert_vm_with_id(vmdto, record['vmid'])
            users_perms = get_list_permissions_of_vm(record['vmid'])
            if users_perms is None:
                continue
            user_model = select_user_db(str(users_perms['ugid']).split('@')[0])
            if user_model is None:
                continue
            add_user_to_vm(user_model.user_id, record['vmid'])
    # If sql has VM that no longer exists in VE, remove it
    vm_sql = get_list_db_vm()
    for record in vm_sql:
        if record.id not in vm_id_list_prox:
            remove_vm_db(record.id)


def get_list_vms_proxmox():
    vms = proxmoxAuth.cluster.resources.get(type='vm')
    return vms


def get_list_permissions_of_vm(vmid: int) -> dict:
    path_temp = '/vms/' + str(vmid)
    vm_acl = proxmoxAuth.access.acl.get()
    for perm in vm_acl:
        try:
            if perm['path'] == path_temp:
                return perm
        except KeyError:
            continue
    return None


def count_words(text) -> int:
    words = text.split()
    return len(words)


def bytes_to_gb(bytes_value):
    gb_value = bytes_value / (1024 ** 3)  # 1 GB = 1024^3 bytes
    return round(gb_value, 3)


def generate_password():
    pwo = PasswordGenerator()
    return pwo.generate()


def remove_vm_by_id(vm_id: int) -> HttpResponse:
    try:
        proxmoxAuth.nodes(default_vm_node).qemu(vm_id).post('status/stop')  # Stop vm if its running
        time.sleep(1)  # Wait for vm to stop
        proxmoxAuth.nodes(default_vm_node).qemu(vm_id).delete()  # Delete from proxmox ve
        remove_vm_db(vm_id)  # Delete from DB
        return HttpResponse("VM removed!")
    except Exception as e:
        return HttpResponse({'error': str(e)})








